class Matrix
{
    FLOAT val[4][4];
public:
    Matrix()
    {
        Zero();
    }
    Matrix(FLOAT_ARG ax, FLOAT_ARG ay, FLOAT_ARG az)
    {
        InitRotate(ax, ay, az);
    }
    Matrix(XYZ axis, FLOAT_ARG ang)
    {
        InitAxisRotate(axis, ang);
    }
    void Zero()
    {
        memset(&val, 0, sizeof val);
    }
    void Identity()
    {
        for(unsigned i=0; i<4; ++i)
            for(unsigned j=0; j<4; ++j)
                val[i][j] = (j==i) ? 1.0 : 0.0;
    }
    void InitRotate(FLOAT_ARG ax, FLOAT_ARG ay, FLOAT_ARG az)
    {
        FLOAT tmp[3][4][4];
        
        Identity();

        memset(&tmp, 0, sizeof(tmp));

        tmp[0][1][1]=tmp[0][2][2]=COS(ax); tmp[0][2][1]=-(tmp[0][1][2]=SIN(ax));
        tmp[1][0][0]=tmp[1][2][2]=COS(ay); tmp[1][0][2]=-(tmp[1][2][0]=SIN(ay));
        tmp[2][0][0]=tmp[2][1][1]=COS(az); tmp[2][1][0]=-(tmp[2][0][1]=SIN(az));

        tmp[0][0][0] = 1;
        tmp[0][3][3] = 1;
        tmp[1][1][1] = 1;
        tmp[1][3][3] = 1;
        tmp[2][2][2] = 1;
        tmp[2][3][3] = 1;

        for(int i=0; i<4; i++)
        {
            FLOAT m2[4];

            for(int j=0; j<4; j++)
                m2[j] = tmp[1][i][0] * tmp[0][0][j] +
                        tmp[1][i][1] * tmp[0][1][j] +
                        tmp[1][i][2] * tmp[0][2][j] +
                        tmp[1][i][3] * tmp[0][3][j];

            for(int j=0; j<4; j++)
                val[i][j] = m2[0] * tmp[2][0][j] +
                            m2[1] * tmp[2][1][j] +
                            m2[2] * tmp[2][2][j] +
                            m2[3] * tmp[2][3][j];
        }
    }
    /* Math borrowed from Povray, matrices.c:Compute_Axis_Rotation_Transform */
    void InitAxisRotate(const XYZ &V, FLOAT_ARG ang)
    {
        V.Normalize();
        Identity();

        const FLOAT cosx = COS(ang);
        const FLOAT sinx = SIN(ang);
        
        val[0][0] = V.x * V.x + cosx * (1.0 - V.x * V.x);
        val[0][1] = V.x * V.y * (1.0 - cosx) + V.z * sinx;
        val[0][2] = V.x * V.z * (1.0 - cosx) - V.y * sinx;
        
        val[1][0] = V.x * V.y * (1.0 - cosx) - V.z * sinx;
        val[1][1] = V.y * V.y + cosx * (1.0 - V.y * V.y);
        val[1][2] = V.y * V.z * (1.0 - cosx) + V.x * sinx;
        
        val[2][0] = V.x * V.z * (1.0 - cosx) + V.y * sinx;
        val[2][1] = V.y * V.z * (1.0 - cosx) - V.x * sinx;
        val[2][2] = V.z * V.z + cosx * (1.0 - V.z * V.z);
    }
    
    void Rotate(XYZ &p)
    {
        register FLOAT tx = p.x*val[0][0] + p.y*val[1][0] + p.z*val[2][0] + val[3][0];
        register FLOAT ty = p.x*val[0][1] + p.y*val[1][1] + p.z*val[2][1] + val[3][1];
        p.z               = p.x*val[0][2] + p.y*val[1][2] + p.z*val[2][2] + val[3][2];
        p.y = ty;
        p.x = tx;
    }
    const Matrix operator * (const Matrix &m) const
    {
        Matrix res;
        for(unsigned i=0; i<3; ++i)
            for(unsigned j=0; j<4; ++j)
                for(unsigned k=0; k<4; ++k)
                    res.val[i][j] += val[i][k] * m.val[k][j];
        res.val[3][3] = 1;
        return res;
    }
    const Matrix &operator *= (const Matrix &m)
    {
        return (*this = *this * m);
    }
};
