class XYZ
{
public:
    FLOAT x,y,z;
    XYZ() : x(0),y(0),z(0)
    {
    }
    XYZ(const XYZ &p) : x(p.x), y(p.y), z(p.z)
    {
    }
    
    XYZ(FLOAT_ARG xx,FLOAT_ARG yy,FLOAT_ARG zz) : x(xx),y(yy),z(zz)
    {
    }
    
    const XYZ &operator +=(const XYZ &p) { x+=p.x; y+=p.y; z+=p.z; return *this; }
    const XYZ &operator -=(const XYZ &p) { x-=p.x; y-=p.y; z-=p.z; return *this; }
    const XYZ &operator /=(FLOAT_ARG p) { x/=p; y/=p; z/=p; return *this; }
    const XYZ &operator *=(FLOAT_ARG p) { x*=p; y*=p; z*=p; return *this; }
    const XYZ operator +(const XYZ &p) const { return XYZ(x+p.x, y+p.y, z+p.z); }
    const XYZ operator -(const XYZ &p) const { return XYZ(x-p.x, y-p.y, z-p.z); }
    const XYZ operator -() const { return XYZ(-x, -y, -z); }
    const XYZ operator /(FLOAT_ARG p) const { return XYZ(x/p, y/p, z/p); }
    const XYZ operator *(FLOAT_ARG p) const { return XYZ(x*p, y*p, z*p); }
    const XYZ &operator= (const XYZ &p)
    {
        x = p.x;
        y = p.y;
        z = p.z;
        return *this;
    }
    
    const FLOAT dist(const XYZ &p)const 
    {
        XYZ d(x-p.x, y-p.y, z-p.z);
        return SQRT(d.x*d.x + d.y*d.y + d.z*d.z);
    }
    const FLOAT length() const
    {
        return SQRT(x*x + y*y + z*z);
    }
    
    void Normalize()
    {
    	*this /= length();
    }
    void SetLength(FLOAT_ARG newlen)
    {
    	*this *= (newlen / length());
    }
    
    template<class T>
    inline void Rotate(T &m)
    {
        m.Rotate(*this);
    }
    template<class T>
    const XYZ RotatedBy(T &m) const
    {
        XYZ tmp = *this;
        m.Rotate(tmp);
        return tmp;
    }
    const XYZ RotatedAround(const XYZ &axis, FLOAT_ARG angle) const;
    const XYZ MirroredAround(const XYZ &axis) const
    {
        // This could be replaced with RotatedAround(axis, DEG2RAD(180)),
        // but as SIN(180)=0 and COS(180)=-1, this is simplified a lot.
        // Terribly lot. :)
        XYZ N(axis);
        N.Normalize();
        return N * (DotProduct(N) * 2.0) - *this;
    }
    const XYZ CrossProduct(const XYZ &p)const
    {
        return XYZ(y*p.z - z*p.y,
                   z*p.x - x*p.z,
                   x*p.y - y*p.x);
    }
    const FLOAT DotProduct(const XYZ &p)const
    {
    	return x*p.x + y*p.y + z*p.z;
    }
    
    void print() const
    {
        printf("<%.8"PF", %.8"PF", %.8"PF"> (length %.8"PF")", x,y,z, length());
    }
};
